-- Part of monadic prompt lazy assertions
-- 
module Main where

import MonadicAssertions
import Prelude hiding (gcd)
import Monad (guard)
import System.Random

-- -----------------------------------------------------------------------
-- Assert that list is sorted by comparing neighbouring elements

sorted :: Ord a => Lazy [a] -> Try ()
sorted xs = 
    pNil xs
  <|> (pCons xs >>= \ (_,ys) -> pNil ys)
  <|> (do
    (x,ys) <- pCons xs
    (do (y,_)  <- pCons ys
        check2 (<) x y)
       &&&
     sorted ys)

sortedT :: [Int] -> [Int]
sortedT xs = assert "sorted" sorted xs

{- Variant in paper:
sorted :: Ord a => Lazy [a] -> Assert
sorted xs = pNil xs
        ||| (do (_,ys) <- pCons xs; pNil ys)
        ||| (do (x,ys) <- pCons xs; (y,_) <- pCons ys;
                ((do rx <- tVal x; ry <- tVal y; guard (rx <= ry))
                 &&& sorted ys))
-}

-- -----------------------------------------------------------------------
-- Assert that list is sorted by comparing an element with all
-- later element (quadratic time complexity).

sortedQ :: Ord a => Lazy [a] -> Assert


sortedQ xs =
      pNil xs
  <|> (pCons xs >>= \(_,ys) -> pNil ys)
  <|> (do
       (y,ys) <- pCons xs
       (allA (check2 (<=) y) ys &&& sortedQ ys))

sortedQT :: [Int] -> [Int]
sortedQT xs = assert "sortedQ" sortedQ xs

-- --------------------------------------------------------------------------
-- Assert that given two lists are equal viewed as sets.
-- Adds this as precondition on tupling function.

equalSets :: Eq a => Lazy [a] -> Lazy [a] -> Assert
equalSets xs ys =
  allA (`elemA` ys) xs &> allA (`elemA` xs) ys

elemA :: Eq a => Lazy a -> Lazy [a] -> Assert
elemA x xs = do
  (y,ys) <- pCons xs
  (check2 (==) x y <|> elemA x ys)

equalSetsT :: [Int] -> [Int] ->
              ([Int],[Int])
equalSetsT xs ys =
  assert "equalSets" 
      (fun2 (\x y z -> equalSets x y))
      (,) xs ys

-- -----------------------------------------------------------------------
-- Assert for gcd that result is a factor of both arguments

gcd :: Int -> Int -> Int
gcd n m = assert "result factor of args"
  (fun2 (\x y res -> 
          check2 (\x r -> x `mod` r == 0)
                 x res
      &&& check2 (\y r -> y `mod` r == 0)
                 y res))
    gcd' n m

gcd' :: Int -> Int -> Int
gcd' n m = 
  let r = n `mod` m in
    if r == 0 then m else gcd n r  -- fault: n should be m

-- fails
test4 = gcd 16 9


-- -----------------------------------------------------------------------
-- Assert that inserting into a sorted list preservers sortedness

checkedInsert :: (Ord a, Observe a) => a -> [a] -> [a]
checkedInsert x xs = insert x (assert "insert" sorted xs)

insert :: (Ord a, Observe a) => a -> [a] -> [a]
insert = assert "insert preserves sortedness"
                (fun2 (\ _ ys zs -> sorted ys ==> sorted zs))
                insert'

insert' :: (Ord a, Observe a) => a -> [a] -> [a]
insert' x [] = [x]
insert' x (y:ys) = if x <= y then x:y:ys else y : insert' x ys

i1 = insert 3 [2,4::Int]     -- correct
i2 = insert 3 [2,3,4::Int]   -- fault: sorted is strict (avoid duplicates)
i3 = insert 3 [5,2,4::Int]   -- correct, because precondition violated


-- -----------------------------------------------------------------------

data Tree a = Node (Tree a) a (Tree a) | Empty deriving Show

instance (Observe a) => Observe (Tree a) where
  obs Empty = o0 "Empty" Empty
  obs (Node l x r) = o3 "Node" Node l x r

pEmpty :: Lazy (Tree a) -> Try ()
pEmpty = p0 (\t -> case t of Empty -> True; _ -> False)

pNode :: Lazy (Tree a) -> Try (Lazy (Tree a),Lazy a,Lazy (Tree a))
pNode = p3 (\t -> case t of Node l x r -> Just (l,x,r); _ -> Nothing)

insertTree :: (Ord a, Observe a) => a -> Tree a -> Tree a
insertTree x t = 
  assert "result of insert tree is ordered"
    (fun2 (\_ yt zt -> sortedTree zt))
    insertTree' x t

sortedTree :: Ord a => Lazy (Tree a) -> Try ()
sortedTree t = between t Nothing Nothing

between :: Ord a => Lazy (Tree a) -> Maybe a -> Maybe a -> Try ()
between lt ml mr = pEmpty lt <|> (do
  (ll,lx,lr) <- pNode lt
  x <- pVal lx
  guard (maybe True (<= x) ml)
  guard (maybe True (x <=) mr)
  between ll ml (Just x) &&& between lr (Just x) mr)
  
insertTree' :: (Ord a, Observe a) => a -> Tree a -> Tree a
insertTree' x Empty = Node Empty x Empty
insertTree' x (Node l y r) = 
  if x <= y 
    then Node (insertTree x l) y r
    else Node l y (insertTree x r)

elemTree :: (Ord a) => a -> Tree a -> Bool
elemTree x Empty = False
elemTree x (Node l y r) = 
  (x==y) || (if x<y then elemTree x l else elemTree x r)

uTree :: Tree Int
uTree = Node (Node Empty 2  (Node Empty 5 Empty)) 4 (Node Empty 7 Empty)

sumT :: Tree Int -> Int
sumT Empty = 0
sumT (Node l x r) = sumT l + x + sumT r

main = do
  list <- sequence (replicate 200 (randomRIO (0,1000)))
  print (sumT (foldr insertTree Empty list))

-- -----------------------------------------------------------------------
-- First or second list element is greater than 0

f :: [Int] -> [Int]
f bs = assert "test non-determinism"
  (\xs -> (pCons xs <|> (pCons xs >>= \(y,ys) -> pCons ys)) >>= \(z,_) -> check1 (>0) z) bs
 

-- -----------------------------------------------------------------------
-- Various tests, mostly repeated from above

sameLengths :: Lazy [a] -> Lazy [b] -> Assert
sameLengths l1 l2 = 
    (pNil l1 &> pNil l2)
  <|> (do
    ((_,xs),(_,ys)) <- pCons l1 <&> pCons l2
    sameLengths xs ys)



sortedLazy :: Ord a => Lazy [a] -> Assert
sortedLazy xs = notAssert $ do
  (y,ys) <- anyCons xs
  (z,_) <- anyCons ys
  check2 (>) y z

anyCons :: Lazy [a] -> Try (Lazy a,Lazy [a])
anyCons xs = do
  (y,ys) <- pCons xs
  return (y,ys) <|> anyCons ys

test1 = assert "test" sortedLazy [1,2,3,2::Int]

checkedZip :: (Observe a, Observe b) => [a] -> [b] -> [(a,b)]
checkedZip = 
  assert "zip args of same length" 
    (fun2 (\xs ys res -> sameLengths xs ys)) zip

-- this one does not fail!
-- Equal lengths do not succeed either but that is ok for a function
test2 = checkedZip [1,2,3,4::Int] [4,3,5::Int]


-- works but highlighting misses parts
test3 = assert "test3" (\t -> pTuple t >>= \(xs,ys) -> sameLengths xs ys) 
          ([1,2,3,4::Int],[4,3,5::Int])

test5 = let l = assert "test5" (\t -> pTuple t >>= \ (x,y) -> pNil x >> pCons y >> return ())
               ([]::[Int],[1::Int]) in
          (snd l,fst l)
