-- Part of monadic prompt lazy assertions
-- 
-- Defines abstract types Try and Lazy and their primitive functions.

module Try
  (Try, Assert, Lazy, assert
  ,notAssert, (<|>),(<$>)
  ,pVal,pPrim,p0,p1,p2,p3,pFun)
  where

import Observer
import PosTrie
import Monad
import System.IO.Unsafe
import Data.IORef

infixl 3  <$>
infixr 2  <|>

data Lazy a = L EvalTreeRef a

type SuccCont a = PosSet -> a -> FailCont -> IO ()
type FailCont = PosSet -> IO ()

newtype Try a = Try (SuccCont a -> FailCont -> IO ())

type Assert = Try ()

fromTry :: Try a -> (SuccCont a -> FailCont -> IO ())
fromTry (Try t) = t

success :: PosSet -> a -> Try a
success poss x = Try $ \sc fc -> sc poss x fc

failT :: PosSet -> Try a
failT p = Try (\sc fc -> fc p)


instance Monad Try where
  return x = Try $ \sc fc -> sc emptyPosSet x fc

  (Try t) >>= f = 
    Try $ \sc fc -> t (\p1 x tfc -> 
      fromTry (f x) (\p2 y ffc -> sc (unionPosSets p1 p2) y ffc)
                    (tfc . (unionPosSets p1))) fc

instance MonadPlus Try where
  mzero = Try (\sc fc -> fc emptyPosSet)
  mplus = (<|>)


notAssert :: Try a -> Try ()
notAssert (Try t) = Try $ \sc fc -> t (\p _ sfc -> fc p) (\p -> sc p () fc)


(<|>) :: Try a -> Try a -> Try a
(Try t1) <|> (Try t2) = Try $ \sc fc -> do
  ref <- newIORef Nothing
  t1 sc (second ref fc) >> t2 sc (second ref fc)

-- When executed the second time call fail continuation with
-- union of both position sets.
-- Assume that reference initialised with Nothing.
second :: IORef (Maybe PosSet) -> FailCont -> PosSet -> IO ()
second ref fc p = do
  v <- readIORef ref
  maybe (writeIORef ref (Just p))
        (\p1 -> fc (unionPosSets p p1))
        v

-- concurrent variant of `ap`
(<$>) :: Try (a -> b) -> Try a -> Try b
(Try t1) <$> (Try t2) = Try $ \sc fc -> do
  fRef <- newIORef []
  xRef <- newIORef []
  ref <- newIORef True
  t1 (\fp f ffc -> do
        updateIORef ((fp,f,ffc):) fRef
        xs <- readIORef xRef
        mapM_ (\ (xp,x,xfc) -> sc (unionPosSets fp xp) (f x) (comb ffc xfc))
              xs)
     (\p -> first ref (fc p))
  t2 (\xp x xfc -> do
        updateIORef ((xp,x,xfc):) xRef
        fs <- readIORef fRef
        mapM_ (\ (fp,f,ffc) -> sc (unionPosSets fp xp) (f x) (comb ffc xfc))
              fs)
     (\p -> first ref (fc p)) 

-- is that correct???
comb :: FailCont -> FailCont -> FailCont
comb fc1 fc2 p = fc1 p >> fc2 p


-- When executed the first time call the given action;
-- do nothing if called again.
-- Assume that reference initialised with True.
first :: IORef Bool -> IO () -> IO ()
first ref action = do
  v <- readIORef ref
  if v then writeIORef ref False >> action
       else return ()

updateIORef :: (a -> a) -> IORef a -> IO a
updateIORef f ref = do
  v <- readIORef ref
  writeIORef ref (f v)
  return v

assert :: Observe a => String -> (Lazy a -> Try ()) -> a -> a
assert label p x =  unsafePerformIO (do
  (r,x') <- observe x
  let Try check = p (L r x)
  check (\_ _ _ -> putStrLn ("\nAssertion succeeded: "++label))
        (\p -> do refTree <- readIORef r
                  shownEvalTree <- showWFail p refTree
                  fail ("\nAssertion failed: " ++ label ++ "\n" ++
                        shownEvalTree ++ "\n"))
  return x')


-- primitive lazy parsers


-- this is wrong; should be fully strict
pVal :: Lazy a -> Try a
pVal = pPrim

pPrim :: Lazy a -> Try a
pPrim rx@(L r v) = Try $ \sc fc -> do
  e' <- readIORef r
  case e' of
    Eval pos _ []    -> sc (oneElemPosSet pos) v fc
    Uneval ref -> suspTIO' ref (pPrim rx) sc fc

p0 :: (a -> Bool) -> Lazy a -> Try ()
p0 p rx@(L r v) = Try $ \c fc -> do
  e' <- readIORef r
  case e' of
    Eval pos _ _   -> if p v then c (oneElemPosSet pos) () fc
                             else fc (oneElemPosSet pos)
    Uneval ref -> suspTIO' ref (p0 p rx) c fc

p1 ::  (a -> Maybe b) -> Lazy a -> Try (Lazy b)
p1 check rx@(L r v) = Try $ \c fc -> do
  e' <- readIORef r
  case e' of
    Eval pos _ [eX] -> maybe (fc (oneElemPosSet pos))
                             (\ x -> c (oneElemPosSet pos) (L eX x) fc)
	                     (check v)
    Eval pos _ _ -> fc (oneElemPosSet pos)
    Uneval ref -> suspTIO' ref (p1 check rx) c fc

p2 ::  (a -> Maybe (b,c)) -> Lazy a -> Try (Lazy b,Lazy c)
p2 check rx@(L r v) = Try $ \c fc -> do
  e' <- readIORef r
  case e' of
    Eval pos _ [eX,eY] -> maybe (fc (oneElemPosSet pos))
                                (\ (x,y) -> c (oneElemPosSet pos) 
                                              (L eX x,L eY y) fc)
		                (check v)
    Eval pos _ _ -> fc (oneElemPosSet pos)
    Uneval ref -> suspTIO' ref (p2 check rx) c fc

p3 :: (a -> Maybe (b,c,d)) -> Lazy a -> Try (Lazy b,Lazy c,Lazy d)
p3 check rx@(L r v) = Try $ \c fc -> do
  e' <- readIORef r
  case e' of
    Eval pos _ [eX,eY,eZ] -> 
      maybe (fc (oneElemPosSet pos))
        (\ (x,y,z) -> c (oneElemPosSet pos) (L eX x,L eY y,L eZ z) fc)
	(check v)
    Eval pos _ _ -> fc (oneElemPosSet pos)
    Uneval ref -> suspTIO' ref (p3 check rx) c fc

-- Helper function used above to suspend when a data part is still unevaluated.
suspTIO' :: IORef (IO ()) -> Try a -> SuccCont a -> FailCont -> IO ()
suspTIO' ref (Try newIO) c fc = do
  io <- readIORef ref
  writeIORef ref (io >> newIO c fc)


-- -------------------------------------
-- A function is matched as a lazy list of tuples

pFun ::  Lazy (a -> b) -> Try (Lazy [(a,b)])
pFun (L reF _) = Try $ \c fc -> do
  listEnd <- emptyUnevalRef
  listRef <- newIORef undefined
  c emptyPosSet (L listEnd (unsafePerformIO (readIORef listRef))) fc
  pFun' reF listEnd listRef
  where 
  pFun' :: EvalTreeRef -> EvalTreeRef -> IORef [(a,b)] -> IO ()
  pFun' r listEnd listRef = do
    evalTree <- readIORef r
    case evalTree of
      Uneval r' ->
        suspTIO r' (pFun' r listEnd listRef)
      (Fun p eArg eRes arg res nextFun) -> do
        newListEnd <- emptyUnevalRef
        Uneval ref <- readIORef listEnd
        act <- readIORef ref
        singleMappingRef <- newIORef (Eval (1/:p) "->" [eArg,eRes])
        writeIORef listEnd
          (Eval p ":" [singleMappingRef,newListEnd])
        newListRef <- newIORef undefined
        writeIORef listRef 
          ((coerce arg,coerce res):unsafePerformIO (readIORef newListRef))
        act
        pFun' nextFun newListEnd newListRef

suspTIO :: IORef (IO ()) -> IO () -> IO ()
suspTIO ref newIO = do
  io <- readIORef ref
  writeIORef ref (io >> newIO)


